﻿
/**
 *「Amazonランキングの謎を解く」という本に載っていたモデルをシミュレートしてみた
 * オリジナル: http://web.econ.keio.ac.jp/staff/hattori/amazonj.htm
 *      2013/08/04 by rikunora
 */

using System;
using System.Collections.Generic;
// using System.Net;
using System.Windows;
using System.Windows.Controls;
// using System.Windows.Input;
using System.Windows.Media;
// using System.Windows.Shapes;
using System.Windows.Controls.DataVisualization.Charting;

using jp.takel.PseudoRandom;

namespace AmazonRank01
{
    public partial class MainPage : UserControl
    {
        // 主要パラメータ
        uint N; // 本の総数
        double A;   // ジャンプ率の最小値 = 最も売れない本の注文頻度
        double B;   // 最大ジャンプ率/最小ジャンプ率の比 = 売れる本と売れない本の平等性を表す
        uint ID;    // 着目する本の番号

        List<Book> OrderList = new List<Book>();

        MersenneTwister Rnd; // 乱数

        public MainPage()
        {
            InitializeComponent();
            _init();
        }

        /// <summary>
        /// 初期化、最初の立ち上げ時に１回だけ
        /// </summary>
        private void _init()
        {
            // 変数の初期化
            N = 100;     // とりあえず100冊
            A = 0.005;   // 年間５冊が最低ラインだそうだ
            B = 1.0;     // 元の本に書いてあった値は 0.809
            ID = N / 4;  // デフォルトで上位1/4の本とする

            // 入力テキストに初期値を設定する
            textBox_N.Text = N.ToString();
            textBox_A.Text = A.ToString();
            textBox_B.Text = B.ToString();
            textBox_I.Text = ID.ToString();

            // 乱数の初期化
            uint seed = (uint)Environment.TickCount;
            this.Rnd = new MersenneTwister(seed);
        }

        /// <summary>
        /// シミュレーション開始時の準備
        /// </summary>
        private void _prepare()
        {
            OrderList.Clear();  // リストを空にする

        //  textBlock1.Text = "初期確率\r\n";

            // リストの作成
            for (int i = 1; i <= N; i++) // 1～Nまで
            {
                // ジップの法則に従って売れる確率を定める
                double prob = Math.Pow( A * ((double)N / i), 1 / B);
                Book a_book = new Book(i, prob );
                OrderList.Add(a_book);
                
            //  textBlock1.Text += String.Format("{0} : {1:f10} \r\n", i, prob);
            }
        }

        /// <summary>
        /// テキストボックスから入力値を取り込む
        /// </summary>
        /// <returns>成功すればtrue</returns>
        private bool checkInput()
        {
            try
            {
                N = uint.Parse( textBox_N.Text );
            }
            catch( Exception ){
                MessageBox.Show("Nの入力に誤りがあります");
                return false;
            }

            try{
                A = double.Parse( textBox_A.Text );
            }
            catch( Exception ){
                MessageBox.Show("Aの入力に誤りがあります");
                return false;
            }
            if (A <= 0.0)
            {
                MessageBox.Show("Aは0より大きくしてください");
                return false;
            }

            try
            {
                B = double.Parse(textBox_B.Text);
            }
            catch (Exception)
            {
                MessageBox.Show("Bの入力に誤りがあります");
                return false;
            }
            if (B <= 0.0)
            {
                MessageBox.Show("Bは0より大きくしてください");
                return false;
            }

            try
            {
                ID = uint.Parse(textBox_I.Text);
            }
            catch (Exception)
            {
                MessageBox.Show("IDの入力に誤りがあります");
                return false;
            }
            if (ID > N)
            {
                MessageBox.Show("IDはN以下にしてください");
                return false;
            }

            return true;
        }

        /// <summary>
        /// ボタンを押してシミュレーション開始
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button1_Click(object sender, RoutedEventArgs e)
        {
            if (false == checkInput())
            {   // 入力値に誤りがあれば、そこでおしまい
                return;
            }

            button1.IsEnabled = false; // ボタン入力不可とする
            button1.Content = "on going...";

            _prepare();

            graphBook = OrderList[(int)ID - 1]; // 着目する本を指定
            graphPointList.Clear(); // リストを空にする
            drawGraph(); // 空の状態で一度グラフを描画する

            // タイマーイベント開始
            timer_count = 0;
            CompositionTarget.Rendering += new EventHandler(MainHandler);
        }

        // タイマーカウンター
        uint timer_count = 0;

        /// <summary>
        /// タイマーで呼び出されるメインハンドラー
        /// </summary>
        /// <param name="o"></param>
        /// <param name="sender"></param>
        private void MainHandler(object o, EventArgs sender)
        {
            const int MAX_TIMER_COUNT = 100; // シミュレーション試行回数
            const int INTERVAL = 10;    // この回数だけtimerが回ったら試行を行う

            if (timer_count % INTERVAL == 0)
            {
                oneSale();  // １回の販売ターン

                printOrderList(); // 現在のリストを表示

                // グラフに点を追加する
                graphPointList.Add(
                    new Point { X = timer_count/INTERVAL, Y = OrderList.IndexOf(graphBook) }
                );
            }

            if (++timer_count >= (MAX_TIMER_COUNT * INTERVAL) )
            {
                simEnd();   // シミュレーション終了
            }
        }

        /// <summary>
        /// シミュレーション終了処理
        /// </summary>
        private void simEnd()
        {
            // タイマーイベント終了
            CompositionTarget.Rendering -= new EventHandler(MainHandler);

            button1.Content = "Start!";
            button1.IsEnabled = true; // ボタン入力可能に戻す

            printSoldCount(); // 売上結果を表示する

            drawGraph(); // グラフを描画する
        }

        /// <summary>
        /// リストの中身を表示、上位MAX_COUNTまでを表示する
        /// </summary>
        private void printOrderList()
        {
            const int MAX_COUNT = 100;

            textBlock1.Text = "";

            int count = 0;
            foreach (Book a_book in OrderList)
            {
                textBlock1.Text += String.Format("{0}, ", a_book.getId() );
                if (++count > MAX_COUNT)
                {
                    break;
                }
            }
        }

        // グラフ関連変数
        Book graphBook; // 着目する本
        List<Point> graphPointList = new List<Point>();

        // グラフを描く
        private void drawGraph()
        {
            this.chart1.DataContext = graphPointList;
            (chart1.Series[0] as ScatterSeries).Refresh();
        }

        /// <summary>
        /// 売上結果を表示する。
        /// 表示するときに、リストをIDでソートし直している。破壊的。
        /// </summary>
        private void printSoldCount()
        {
            // リストをIDでソートする
            OrderList.Sort(delegate(Book A, Book B) { return A.getId() - B.getId(); });

            textBlock1.Text = "ID, SoldCount\r\n";

            foreach (Book a_book in OrderList)
            {
                textBlock1.Text += String.Format("{0}, {1}\r\n", a_book.getId(), a_book.count);
            }
        }

        /// <summary>
        /// １回の販売ターン
        /// </summary>
        private void oneSale()
        {
            // 今回売れた本のリスト
            List<Book> soldList = new List<Book>();

            foreach (Book a_book in OrderList)
            {
                // それぞれの本が固有の確率で売れる
                if ( a_book.getPJump() >= this.Rnd.NextDouble() )
                {
                //  MessageBox.Show( String.Format( "{0} が売れた!", a_book.getId() ) );
                    a_book.count++; // 売上カウントアップ
                    soldList.Add(a_book);   // 売れたリストに追加する
                }
            }

            // 売れた本を先頭に持ってゆく
            foreach (Book s_book in soldList)
            {
                OrderList.Remove(s_book);
                OrderList.Insert(0, s_book);
            }

            soldList.Clear(); // 用済み
        }
    }

    /**
     * 書籍オブジェクト
     */
    class Book
    {
        int id; // 本のID番号
        double p_jump;  // ジャンプ率

        public int count;   // 売上数

        public Book(int id0, double p_jump0)
        {
            this.id = id0;
            this.p_jump = p_jump0;
            this.count = 0;
        }

        public int getId()
        {
            return this.id;
        }

        public double getPJump()
        {
            return this.p_jump;
        }
    }
}
